/* Daily routine monitor for Arduino Mega */
/* CC BY-NC-SA Jeroen Brinkman */
/* Version 4 dd 1-1-2024 */

#include <Arduino.h>
#include <TM1637Display.h>
#include "EEPROM.h"

#define BOUNCE      8   // Number of debouncing samples
#define ZERO        0   // Place where to start with storing in eprom. Must be 0, 20, 40 etc
#define PRESS    2000   // Milliseconds to press to detect a long press
#define STORE  600000   // Milliseconds to wait before storing the settings in EEPROM (= 10 minutes)
#define BLINK     500   // Milliseconds used for blinking
#define ON          0   // Milliseconds used for blinking
#define OFF         1   // Milliseconds used for blinking

#define MONTH       1   // Show this month
#define YEAR        2   // Show this year
#define SUM        18   // Show sums text
#define PERCENT    19   // Show percentages text
#define BRIGHT     15   // Show brightness text

#define WHITE      13   // EEPROM position where the brightness white display is stored
#define GREEN      14   // EEPROM position where the brightness green display is stored
#define RANGE      15   // EEPROM position where the range is stored
#define SHOW       16   // EEPROM position where the show is stored
#define LEAP       17   // EEPROM position where the leapyear is stored

#define CLKW      4     // White display clock
#define DIOW      5     // White display data
#define CLKG      6     // Green display clock
#define DIOG      7     // Green display data
#define ROTA      9     // Use ROTB 8 if right turn does not increase month
#define ROTB      8     // Use ROTA 9 if right turn does not increase month
#define ROTP     10     // Rotary push button 
#define PUSH     11     // Normal push button
#define PUSHLED  12     // Green LED from the push button
#define HEART    13     // blinking led showing the programma's hartbeat
#define D29      14     // Day 29 active
#define D30      15     // Day 30 active
#define D31      16     // Day 31 active


// Create display objects
TM1637Display displayW(CLKW, DIOW); // define white display
TM1637Display displayG(CLKG, DIOG); // define green display

int brightW, brightG, buttons, i, j, temp, status, show, todo, range, signalp, signalr, signalA, signalB, total, rot, passed, actual;
unsigned long tostore, pressed;
byte punt;
bool stored, new_signal, leap;

//                    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31
const int days[32] = {0,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47}; // the pin number for the corresponding day switch
// Use the solder check program to check and adapt the sequence of the above pin numbering.

const int months[13] = {0,31,28,31,30,31,30,31,31,30,31,30,31}; // number of days per month
const int cummulative[13] = {0,31,59,90,120,151,181,212,243,273,304,334,365}; // cummulative days per month


// Dutch text
const uint8_t txt[20][4] = {{0x00,0x00,0xBF,0x3F},{0x00,0x0E,0x77,0x37},{0x00,0x71,0x79,0x7C},{0x33,0x27,0x50,0x78},{0x00,0x5F,0x73,0x31},
{0x33,0x27,0x79,0x30},{0x00,0x0E,0x1C,0x54},{0x00,0x0E,0x1C,0x30},{0x00,0x5F,0x3E,0x3D},{0x00,0x6D,0x79,0x73},{0x00,0x3F,0x58,0x78},
{0x00,0x37,0x5C,0x1C},{0x00,0x5E,0x79,0x39},{0x00,0x00,0x00,0x00},{0x00,0x00,0x00,0x00},{0x76,0x79,0x38,0x5E},{0x40,0x5B,0x7F,0x40},
{0x40,0x5B,0x6F,0x40},{0x6D,0x40,0x00,0x00},{0x73,0x40,0x00,0x00}};

/* English text
const uint8_t txt[20][4] = {{0x00,0x00,0xBF,0x3F},{0x00,0x0E,0x77,0x37},{0x00,0x71,0x79,0x7C},{0x33,0x27,0x5F,0x50},{0x00,0x5F,0x73,0x31},
{0x33,0x27,0x5F,0x6E},{0x00,0x0E,0x1C,0x54},{0x00,0x0E,0x1C,0x30},{0x00,0x5F,0x3E,0x3D},{0x00,0x6D,0x79,0x73},{0x00,0x3F,0x58,0x78},
{0x00,0x37,0x5C,0x1C},{0x00,0x5E,0x79,0x39},{0x00,0x00,0x00,0x00},{0x00,0x00,0x00,0x00},{0x7C,0x50,0x10,0x00},{0x40,0x5B,0x7F,0x40},
{0x40,0x5B,0x6F,0x40},{0x6D,0x40,0x00,0x00},{0x73,0x40,0x00,0x00}};

*/

/* Important variables

actual is the shown month
total is the number of daily routines of the shown month
passed is the number of days passed up to the last daily routine

todo	
0	Nothing happened
1	Push button pressed      -> toggle sums/percentages
2	Long press push button   -> write totals to EEPROM and than blink the LED of the button
3	Rotate button pressed    -> set the brightness form the white and the green display
4	Long press rotate button -> toggle month view/year view
5	Rotate to the right      -> increase the month/brightness
6	Rotate to the left       -> decrease the month/brightness
7   STORE time has passed    -> Update settings in EEPROM 

show -> short press push button
1	SUM - Show sums
2	PERCENT - Show percentages

range -> long press push button
1   MONTH - Display this month
2   YEAR - Display the total up to and including this month

status -> short press rotary button
1	SUM - Normal operation with sums
2	PERCENT - Normal operation with percentages
3	WHITE - Set brightness white display
4	GREEN - Set brightness green display
5   YEAR - Define weather this year is a leapyear

Content text array 
 0> 0.0<
 1> JaN<
 2> FEb<
 3>MmaR<
 4> Apr<
 5>MmaY<
 6> Jun<
 7> Jul<
 8> AUG<
 9> SEp<
10> Oct<
11> Nov<
12> DEC<
13>    <
14>    <
15>bri <
16>-28-<
17>-29-<
18>S-  <
19>P-  <

Content EEPROM
 0 - the month represented by the pressed buttons 
 1 - number of performed daily routines in January
 2 - number of performed daily routines in February
 3 - number of performed daily routines in March
 4 - number of performed daily routines in April
 5 - number of performed daily routines in May
 6 - number of performed daily routines in June
 7 - number of performed daily routines in July
 8 - number of performed daily routines in August  
 9 - number of performed daily routines in September
10 - number of performed daily routines in October
11 - number of performed daily routines in November
12 - number of performed daily routines in December
13 - White brightness (1-7) (stored after a period of time)
14 - Green brightness (1-7) (stored after a period of time)
15 - range (MONTH or YEAR)  (stored after a period of time)
16 - show (sums or percentages) (stored after a period of time)
17 - leapyear (true or false)
*/

int len(int val) {
  int length = 0;
  if (val < 10) length = 1;
  if ((val > 9) && (val < 100)) length = 2;
  if ((val > 99) && (val < 1000)) length = 3;
  if (val > 999) length = 4;
  return length;
}

void setup() {
  

  /* define I/O  */
  pinMode(ROTA, INPUT_PULLUP); 
  pinMode(ROTB, INPUT_PULLUP);
  pinMode(ROTP, INPUT_PULLUP); 
  pinMode(PUSH, INPUT_PULLUP);
  pinMode(D29, OUTPUT);
  pinMode(D30, OUTPUT);
  pinMode(D31, OUTPUT);
  pinMode(HEART, OUTPUT); // blinking led showing the programma's hartbeat
  pinMode(PUSHLED, OUTPUT); // led from the push button
  digitalWrite(PUSHLED, ON);
  for (i = 1; i < 32; i++) pinMode(days[i], INPUT_PULLUP); // define all day buttons

  /* Set the brightness  */
  if ((EEPROM.read(ZERO+WHITE) > 7) || (EEPROM.read(ZERO+WHITE) == 0)) EEPROM.write(ZERO+WHITE, 7); // set the white brightness to max if first time
  if ((EEPROM.read(ZERO+GREEN) > 7) || (EEPROM.read(ZERO+GREEN) == 0)) EEPROM.write(ZERO+GREEN, 7); // set the green brightness to max if first time
  brightW = EEPROM.read(ZERO+WHITE); displayW.setBrightness(brightW, true);
  brightG = EEPROM.read(ZERO+GREEN); displayG.setBrightness(brightG, true);

  /* Check the EEPROM data  */
  if ((EEPROM.read(ZERO) > 12) || (EEPROM.read(ZERO) < 1)) EEPROM.write(ZERO, 1); // set the initial month january if out of range
  for (i = 1; i < 32; i++) { 
    if (EEPROM.read(ZERO+i) > months[i]) EEPROM.write(ZERO+i, 0); // reset the number of events if out of range
  }
  if ((EEPROM.read(ZERO+RANGE) != MONTH) && (EEPROM.read(ZERO+RANGE) != YEAR)) EEPROM.write(ZERO+RANGE, MONTH); // set the range to month if first time
  if ((EEPROM.read(ZERO+SHOW) != SUM) && (EEPROM.read(ZERO+SHOW) != PERCENT)) EEPROM.write(ZERO+SHOW, SUM); // set the show to sum if first time
  if ((EEPROM.read(ZERO+LEAP) != 0) && (EEPROM.read(ZERO+LEAP) != 1)) EEPROM.write(ZERO+LEAP, 0); // set the leapyear to false

  /* initialize variables  */
  actual = EEPROM.read(ZERO);
  range = EEPROM.read(ZERO+RANGE); 
  show = EEPROM.read(ZERO+SHOW); 
  leap = EEPROM.read(ZERO+LEAP); 

  status = show;
  stored = true;
  tostore = 0;
}

void loop() {
  digitalWrite(HEART, (millis() / 1000) % 2); //1s heartbeat for the onboard led
 
  /* read the day buttons */
  buttons = 0;
  for (i = months[EEPROM.read(ZERO)]; i > 0; i--) buttons += !digitalRead(days[i]); // number of pressed buttons for the button month

  /* read control buttons and debounce */

  signalp = 0; signalr = 0; rot = 0;
  for (int i = 0; i <= BOUNCE; i++) { // eliminate bouncing
    signalp += !digitalRead(PUSH);
    signalr += !digitalRead(ROTP);
    signalA = digitalRead(ROTA);
    signalB = digitalRead(ROTB);
    if (signalA > signalB) rot -= 1;
    if (signalA < signalB) rot += 1;
    delay(1);
  }
  signalp = (signalp > (BOUNCE / 2)) ? HIGH : LOW;
  signalr = (signalr > (BOUNCE / 2)) ? HIGH : LOW;
  todo = 0;
  
  /* check push button */
  if (signalp == HIGH) {
    pressed = millis() + PRESS;
    digitalWrite(PUSHLED, OFF); // Push button led off
    while (digitalRead(PUSH) == LOW){
      if (millis() > pressed) digitalWrite(PUSHLED, ON); // Push button led on
    }
    digitalWrite(PUSHLED, ON); // Push button led on
    if (millis() < pressed) {
      todo = 1;  
    } else {
      todo = 2;  
    }
  }

  /* check rotary push */
  if (signalr == HIGH) {
    pressed = millis() + PRESS;
    while (digitalRead(ROTP) == LOW) {
      if (millis() > pressed) digitalWrite(PUSHLED, OFF); // Push button led on
    }
    digitalWrite(PUSHLED, ON); // Push button led on
     if (millis() < pressed) {
      todo = 3;  
    } else {
      todo = 4;  
    }
  }
  
  if (rot == 0) {
    new_signal = true;
  } else {
    if ((rot < 0) && new_signal) todo = 5;
    if ((rot > 0) && new_signal) todo = 6;
    new_signal = false;
  }

 /* ------- Determine when to store the settings ------- */
  if (todo == 0) { // when nothing has been pressed for STORE milliseconds, save the settings in EEPROM
    if ((millis() > tostore) && (!stored)) todo = 7;
      if (EEPROM.read(ZERO+WHITE) == brightW) todo = 0; 
      if (EEPROM.read(ZERO+GREEN) == brightG) todo = 0; 
      if (EEPROM.read(ZERO+RANGE) == range) todo = 0; 
      if (EEPROM.read(ZERO+SHOW) == show) todo = 0; 
      if (EEPROM.read(ZERO+LEAP) == leap) todo = 0;
      if (todo == 0) stored = true;
  } else { // tostore != 0
    tostore = millis() + STORE;
    stored = false;
  }

 /* ------- Process todo ------- */
  switch (todo) {
    case 1: // Push button pressed -> toggle sums/percentages
      if (show == SUM) {
        show = PERCENT;
        status = PERCENT;
      } else {
        show = SUM;
        status = SUM;
      }
      break; 
    case 2: // Long press push button -> toggle month view/year view
      if (range == MONTH) {
        range = YEAR;
      } else {
        range = MONTH;
      }
      break;
    case 3: // Rotate button pressed -> set the brightness for the white and the green display and select leapyear
      switch (status) {
        case PERCENT:
          status = WHITE;
          break;
        case SUM:
          status = WHITE;
          break;
        case WHITE:
          status = GREEN;
          break;   
        case GREEN: 
          status = LEAP;
          break;
        case LEAP: 
          status = show;
          break;  
      }
      break;  
    case 4: // Long press rotary button -> write totals to EEPROM and than blink the LED of the push button
      EEPROM.update(ZERO+EEPROM.read(ZERO), buttons); 
      i = EEPROM.read(ZERO) + 1;
      if (i == 13) i = 1;
      EEPROM.update(ZERO, i);
      actual = i;
      break;   
    case 5: // Rotate to the right -> increase the month/brightness
     switch (status) {
        case WHITE:
          brightW += 1; if (brightW == 8) brightW = 7;
          displayW.setBrightness(brightW, true);
          break;   
        case GREEN: 
          brightG += 1; if (brightG == 8) brightG = 7;
          displayG.setBrightness(brightG, true);
          break;
        case LEAP: 
          leap = !leap; 
          break;  
        default:
          actual += 1; if (actual == 13) actual = 1;
          break; 
      }
      break; 
    case 6: // Rotate to the left -> decrease the month/brightness   
     if (status == WHITE) {
        brightW -= 1; if (brightW == 0) brightW = 1;
        displayW.setBrightness(brightW, true);
      } else {
        if (status == GREEN) {
          brightG -= 1; if (brightG == 0) brightG = 1;
          displayG.setBrightness(brightG, true);
        } else {
          actual -= 1; if (actual == 0) actual = 12;
        }
      }
      break; 
    case 7: // Update setting in EEPROM  
      if (!stored) { 
        EEPROM.update(ZERO+WHITE, brightW); 
        EEPROM.update(ZERO+GREEN, brightG); 
        EEPROM.update(ZERO+RANGE, range); 
        EEPROM.update(ZERO+SHOW, show); 
        EEPROM.update(ZERO+LEAP, leap); 
        digitalWrite(PUSHLED, 1); delay(BLINK); digitalWrite(PUSHLED, 0); delay(BLINK); 
        digitalWrite(PUSHLED, 1); delay(BLINK); digitalWrite(PUSHLED, 0); delay(BLINK); 
        digitalWrite(PUSHLED, 1); delay(BLINK); digitalWrite(PUSHLED, 0); delay(BLINK); 
        digitalWrite(PUSHLED, 1); delay(BLINK); digitalWrite(PUSHLED, 0); 
      }
      stored = true;
      break; 
  }

/* ------- Set LEDs buttons 28-31 ------- */
  switch (months[actual]) {
    case 28:// february
      digitalWrite(D29, !leap);     
      digitalWrite(D30, 1); 
      digitalWrite(D31, 1); 
      break; 
    case 30:// 30 days in a month
      digitalWrite(D29, 0); 
      digitalWrite(D30, 0); 
      digitalWrite(D31, 1); 
      break; 
    case 31:// 31 days in a month
      digitalWrite(D29, 0); 
      digitalWrite(D30, 0); 
      digitalWrite(D31, 0);
      break;     
  }


  /* ------- calculate total and passed ------- */

  if (actual < EEPROM.read(ZERO)) total = EEPROM.read(ZERO + actual);
  if (actual == EEPROM.read(ZERO)) total = buttons;
  if (actual > EEPROM.read(ZERO)) total = 0;

  if (actual == EEPROM.read(ZERO)) { // if actual is the buttonized month
    for (passed = months[actual]; digitalRead(days[passed]); passed--); //determine the lastly executed daily routine
  } else { 
    passed = months[actual];
  }
  
  if (range == YEAR) { // display the sum/mean of all months
    for (i = 1; i < actual; i++) total += EEPROM.read(ZERO+i); // adds the totals from january until actual to total
    passed += cummulative[actual-1]; // add days in the previous months to passed
      if (actual > EEPROM.read(ZERO)) passed = 0;
  }

  /* ------- write displays ------- */
  temp = show;
  if ((status == WHITE) || (status == LEAP)) temp = status;
  switch (temp) { // write white display
    case SUM:// Normal operation with sums
      if (range == YEAR) {
        displayW.setSegments(txt[SUM], 2, 0);  // display shows left "S-" 
        displayW.showNumberDec(actual, false, 2, 2); // display shows right the actual month number
      } else {
        displayW.setSegments(txt[actual], 4, 0); // display shows the actual month text
      }
      break; 
    case PERCENT:// Normal operation with percentages
      if (range == YEAR) {
        displayW.setSegments(txt[PERCENT], 2, 0); // display left "P-"   
        displayW.showNumberDec(actual, false, 2, 2);    // display shows right the actual month number
      } else {
        displayW.setSegments(txt[actual], 4, 0); // display shows the actual month text
      }
      break;
    case WHITE:// Set brightness white display
      displayW.setSegments(txt[BRIGHT], 4, 0); // display shows BRIGHT text 
      break;   
    case LEAP:// Set leapyear
      displayW.setSegments(txt[2], 4, 0); // display shows february 
      break;   
  }

  temp = show;
  if ((status == GREEN) || (status == LEAP)) temp = status;
  switch (temp) { // write green display
    case SUM:// Normal operation with sums
       displayG.showNumberDec(total, false);
       break; 
    case PERCENT:// Normal operation with percentages
      i = 0; 
      if (passed != 0) i = (total*1000)/passed; // show one digit behind the comma
      j = len(i); 
      if (j == 1) punt = 0b10000000;
      if (j == 2) punt = 0b01000000;
      if (j == 3) punt = 0b00100000; 
      if (j == 4) punt = 0b00100000;
      if (i==0) { 
        displayG.setSegments(txt[ZERO], 4, 0);
      } else {
        displayG.showNumberDecEx(i, punt, false);
      }
      break;  
    case GREEN:// Set brightness green display
      displayG.setSegments(txt[BRIGHT], 4, 0); // display shows BRIGHT text 
      break;  
   case LEAP:// Set leapyear
      displayG.setSegments(txt[LEAP-1+leap], 4, 0); // display shows -28- or -29- 
      break;   
  }
}